
/*********************************************************************

	cimage_sv: A standalone 3D surface viewer of single slices
	for the Crystal Image package. 
	Place binary in /usr/local/cimage/bin

*********************************************************************/






/*		MODULE FOR CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


/* NOTE: For the port to gtkglext, check out this:

http://www.mail-archive.com/gtk-app-devel-list@gnome.org/msg10995.html

*/


/* for GL, we need GL and GLU. For the interface with gtk,
   we need gtkglarea, a separate module --- scratch this, we need  gtkglext-1.0 */

/* New gtkgl-ext */
#include	<gtk/gtkgl.h>
#include	<gdk/gdkgl.h>
//#include	<gtkgl/gtkglarea.h>			Old glarea
#include	<GL/gl.h>
#include	<GL/glu.h>

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>



/* Stuff for gtk and Imager in general */

#include "../src/gtkhelpers.h"
#include "../src/defs.h"
#include "../src/tiffstuff.h"
#include "../src/imgbuf.h"
#include "../src/dlgbox.h"
#include "../src/functions.h"
#include "../src/support.h"



/**********************************************************

	3D viewers and other OpenGL-based stuff

**********************************************************/

typedef struct 
{
	float red;
	float green;
	float blue;
}
rgbftriplet;


GtkWidget *statusbar1;
static int statusbar_was_used = 0;
int current_lut;



/* Start with the mundane... */


/* Print usage help */

void prhelp ()
{
	dp (0, "imager_sv: Standalone 3D surface viewer for Crystal Image\n");
	dp (0, "(c) 2010 Mark A. Haidekker\n");
	dp (0, "This program should be called from Crystal Image but can be used\nindependently through command-line arguments.\n");
	
	dp (0, "valid options are:\n");
	dp (0, "-l   --load        filename: Specify the filename to load/display\n");
	dp (0, "-t   --tempdir     directory: Specify location of the file\n");
	dp (0, "-d   --debuglevel  level: Specify debug message verbosity (0 - 3)\n");
	dp (0, "-h   --help        : print this help message\n");
	exit (-1);	/* Return an error code bceause this exit indicates failure for call from program */
}


/* This callback function is called when a pixbuf is referenced to zero. It is used to
	free the memory and set the pointer to zero */

void pixbuf_destroy_handler (guchar *pixelptr, void *user_data)
{
GdkPixbuf *pixbuf;

	pixbuf = (GdkPixbuf*)user_data;		/* A pointer to the pixbuf we are dealing with */
	dp (2, "Destroying pixbuf %08X, data at %08X\n",pixbuf, pixelptr);
	if (pixelptr)
	{
		free (pixelptr);
		pixelptr = NULL;
		user_data = NULL;
	}
}


/*	Display a text message in the control window's status bar */

void statusbar_display (const char* text)
{

guint context;

	context = gtk_statusbar_get_context_id((GtkStatusbar*)statusbar1, "mycontext");

	if (statusbar_was_used) /* pop the last message */
	{
		gtk_statusbar_pop((GtkStatusbar*)statusbar1, context);
		statusbar_was_used=0;
	}
	
	gtk_statusbar_push((GtkStatusbar*)statusbar1, context, text);
	statusbar_was_used = 1;
}



/* Run through the entire image and extract min and max values */

void make_minmax (image_type* img)
{
int min,max,i;
double fmin,fmax;
int bitmask;
long maxelems;

	min=0, max=255; bitmask=0; fmin=0; fmax=0;
	maxelems = img->xmax * img->ymax * img->zmax;
	if (img->imgtype == B_RGB) maxelems *= 3;
	switch (img->imgtype)
	{
		case B_CHAR:
		case B_RGB:
		{
			unsigned char* p = img->data;
			min = max = *p;			/* initialize min and max */
			for (i=0; i< maxelems; i++)
			{
				min = MIN (*p,min); max = MAX (*p,max);
				bitmask |= *p;
				p++;
			}
			img->minval.imin=min;
			img->maxval.imax=max;
			break;
		}
		case B_SHORT:
		{
			short* p = img->data;
			min = max = *p;			/* initialize min and max */
			for (i=0; i< maxelems; i++)
			{
				min = MIN (*p,min); max = MAX (*p,max);
				bitmask |= *p;
				p++;
			}
			img->minval.imin=min;
			img->maxval.imax=max;
			break;
		}
		case B_FLOAT:
		{
			float* p = img->data;
			fmin = fmax = *p;			/* initialize min and max */
			for (i=0; i< maxelems; i++)
			{
				if (fmin>*p) fmin=*p;
				if (fmax<*p) fmax=*p;
				p++;
			}
			img->minval.fmin=fmin;
			img->maxval.fmax=fmax;
			break;
		}
	}

	if (img->imgtype==B_FLOAT)
		dp (2, "Done reading. Min-Max are %f ... %f\n",
						fmin,fmax);
	else
		dp (2, "Done reading. Min-Max are %d ... %d, bitmask %04x\n",
					min,max, bitmask);
}
	
	
	
/*	Save a file through imlib (always 3-byte RGB format) */



int save_file (char* filename, GdkPixbuf *my_im)
{
gint success;

	/* If there is no image loaded, exit */
	
	if (!my_im) return -1;

    /* Now save the image specified as the first argument.
		Returns 0 if error, 1 if success */
	
	success=0;
	if (!success) 
	{
		return -1;
	}
	return 0;	
	
}



void mirrory (image_type *img)		/* Display is mirrored about Y, so un-mirror it */
{
int x,y,y1;


	for (x=0; x<img->xmax; x++)
	{
		for (y=0; y<img->ymax/2-1; y++)		/* with odd y, the middle column is left out */
		{
			y1 = img->ymax-y-1;
			switch (img->imgtype)
			{
				case B_CHAR:
				{
					unsigned char *p = img->data;
					unsigned char c;
					c = p[x + img->xmax*y]; p[x + img->xmax*y] = p[x + img->xmax*y1]; p[x + img->xmax*y1] = c;
					break;
				}
				case B_SHORT:
				{
					short *p = img->data;
					short c;
					c = p[x + img->xmax*y]; p[x + img->xmax*y] = p[x + img->xmax*y1]; p[x + img->xmax*y1] = c;
					break;
				}
				case B_FLOAT:
				{
					float *p = img->data;
					float c;
					c = p[x + img->xmax*y]; p[x + img->xmax*y] = p[x + img->xmax*y1]; p[x + img->xmax*y1] = c;
					break;
				}
				case B_RGB:
				{
					rgbtriplet *p = img->data;
					rgbtriplet c;
					c = p[x + img->xmax*y]; p[x + img->xmax*y] = p[x + img->xmax*y1]; p[x + img->xmax*y1] = c;
					break;
				}
			}
		}
	}

}



/* These two dummy functions satisfy dependencies */

void set_display_title (const char *title)
{

}

void ifft2d (image_type *fftimg)
{

}



/****************************************************+

	L O O K - U P   T A B L E S
	
*****************************************************/



/* Color look-up tables. Grayval is scaled between 0 and 1.
   The values returned are between 0 and 1 as well (GL convention) */



rgbftriplet predef_lut_default (float grayval)
{
rgbftriplet color;

	color.red = color.green = color.blue = grayval;
	return color;
}


rgbftriplet predef_lut_rgb (float grayval)
{
rgbftriplet color;

	if ((grayval<=0.0) || (grayval > 1.0))
	{
		color.red=0.0; color.green=0.0; color.blue=0.0;
	}
	else if (grayval < 0.5)
	{
		color.red=0;
		color.blue=1.0-2.0*grayval;
		color.green=2.0*grayval;
	}
	else
	{
		color.blue=0;
		color.green=2.0-2.0*grayval;
		color.red=2.0*(grayval-0.5);
	}

	return color;
}


rgbftriplet predef_lut_rgbk (float grayval)
{
rgbftriplet color;

	if (grayval < 1e-3)
	{
		color.red = color.green = color.blue = 0.0;
		return color;
	}
	else
		return predef_lut_rgb (grayval);
}



/* A close relative of the simple RGB LUT, this LUT
   runs along the brim of the HSV circle and thus covers all hues
   from 0 degrees (red) to 300 degrees (pink) at full saturation 
   and brightness */

rgbftriplet predef_lut_rainbow (float grayval)
{
rgbftriplet rgbf;


	if (grayval < 0.25)
	{
		rgbf.blue=1.0;
		rgbf.green = 4*grayval;
		rgbf.red=0;
	}
	else if (grayval < 0.5)
	{
		rgbf.blue=2-4*grayval;
		rgbf.green = 1;
		rgbf.red= 0;
	}
	else if (grayval < 0.75)
	{
		rgbf.blue=0;
		rgbf.green = 1;
		rgbf.red= 4*grayval-2;
	}
	else
	{
		rgbf.blue=0;
		rgbf.green = 4-4*grayval;
		rgbf.red= 1;
	}

	return rgbf;

}


/* A great general-purpose LUT "lifted" from NIH image,
   it shows dark shades in hues of black and blue, then
   progresses over magenta, red, orange, yellow
   to white. */

rgbftriplet predef_lut_fire (float grayval)
{
rgbftriplet color;
int i;

	color.red = color.green = color.blue = 0.0;
	i = (int)(256*grayval);
		
	if (i<64) color.blue=4*i;
	else if (i<128) color.blue=511-4*i;
	else if (i>212) color.blue=6*i-1275;
		
	if ((i>=16) && (i<=143)) color.red=2*i-32;
	else if (i>143) color.red=255;
		
	if ((i>100) && (i<=227)) color.green=2*i-200;
	else if (i>227) color.green=255;

	color.red /= 256;
	color.green /= 256;
	color.blue /= 256;

	return color;
}



/* A Gamma LUT with gamma=2 to enhance contrast in the dark regions */

rgbftriplet predef_lut_gamma (float grayval)	/* Gamma of two is sqrt */
{
	return predef_lut_default (sqrt (grayval));
}


/* The main CLUT function */

rgbftriplet surf3d_clut (float grayval)
{

	switch (current_lut)
	{
		case 0:											/* Default */
		case 3:											/* Specified RGB */
			return predef_lut_rgb (grayval);
		case 1:											/* Grayscale */
			return predef_lut_default (grayval);
		case 2:											/* Negative Grayscale */
			return predef_lut_default (1.0-grayval);
		case 4:											/* NIH Image Fire LUT */
			return predef_lut_rgbk (grayval);
		case 5:											/* NIH Image Fire LUT */
			return predef_lut_fire (grayval);
		case 6:											/* Rainbow (run through hue in HSV space) */
			return predef_lut_rainbow (grayval);
		case 10:										/* Grayscale gamma=2.0 */
			return predef_lut_gamma (grayval);

		default:
			return predef_lut_default (grayval);
	}
}



/****************************************************+

	3D surface plot
	
*****************************************************/



GtkWidget* surf3dwnd;
GtkWidget* surf3d_awnd;
GtkAdjustment* surf3d_azimuth_adj;
GtkAdjustment* surf3d_elev_adj;
GtkAdjustment* surf3d_zscale_adj;
GtkAdjustment* surf3d_subsamp_adj;
GtkAdjustment* surf3d_perspective_adj;

/* Globals defs as of 0.9.0 - do they reall need to be global? */

GdkGLContext *glcontext = NULL;
GdkGLDrawable *gldrawable = NULL;
GdkGLConfig  *glconfig = NULL;

GtkWidget	*glsurfarea=NULL;
int gl_widget_ready;

image_type surf3d_img;

/* Pointers to hold:
	1) The vertices (3 floats)
	2) The respective color (3 floats)
	3) A normal vector (3 floats)
*/

static GLfloat* surf3d_vert = NULL;
static GLfloat* surf3d_color = NULL;
static GLfloat* surf3d_norm = NULL;

/* Global stuff */

static int surf3d_vxmax=0;
static int surf3d_vymax=0;
static int surf3d_prepared = 0;
static int surf3d_awnd_exists = 0;
static int surf3d_wnd_exists = 0;
static int surf3d_bgcolor = 0;
unsigned char light_en;
unsigned char surf3d_spotlight;
unsigned char surf3d_showlight;

float surf3d_lv_elev, surf3d_lv_azim, surf3d_lv_spec;
float surf3d_eyedist, surf3d_lvdist;
float surf3d_perspective;



/* Data preparation */

/* Compute the normalized (len=1) cross product of two vectors v1 and v2 */

void crossprodf (float v1[3], float v2[3], float *res)
{
float len;

	res[0] = v1[1]*v2[2] - v1[2]*v2[1];
	res[1] = v1[2]*v2[0] - v1[0]*v2[2];
	res[2] = v1[0]*v2[1] - v1[1]*v2[0];

	len = sqrt (SQR(res[0]) + SQR(res[1]) + SQR(res[2]));
	if (len==0.0)
	{
		res[0]=0.0; res[1]=0.0; res[2]=1.0; /* Generic default */
	}
	else
	{
		res[0] /= len; res[1] /= len; res[2] /= len;
	}

}

/* Another cross product, but this one takes three CCW (!) vertices
   that form a triangle */

void crossprodf2 (float v1[3], float v2[3], float v3[3], float *res)
{
float d1[3], d2[3];
int i;

	for (i=0; i<3; i++)
	{
		d1[i]=v2[i]-v1[i]; d2[i]=v3[i]-v2[i];
	}
	crossprodf (d1, d2, res);
}




/*** OPEN GL SECTION ***/

void cube_at (float x, float y, float z)
{
float d = 0.03;

	glColor4f (1.0,0.0,0.0,1.0);
	
	glBegin (GL_QUADS);
	glVertex3f (x-d, y-d, z+d);
	glVertex3f (x+d, y-d, z+d);
	glVertex3f (x+d, y+d, z+d);
	glVertex3f (x-d, y+d, z+d);

	glVertex3f (x-d, y-d, z-d);
	glVertex3f (x+d, y-d, z-d);
	glVertex3f (x+d, y+d, z-d);
	glVertex3f (x-d, y+d, z-d);

	glVertex3f (x-d, y+d, z+d);
	glVertex3f (x+d, y+d, z+d);
	glVertex3f (x+d, y+d, z-d);
	glVertex3f (x-d, y+d, z-d);

	glVertex3f (x-d, y-d, z+d);
	glVertex3f (x+d, y-d, z+d);
	glVertex3f (x+d, y-d, z-d);
	glVertex3f (x-d, y-d, z-d);

	glVertex3f (x+d, y+d, z+d);
	glVertex3f (x+d, y-d, z+d);
	glVertex3f (x+d, y-d, z-d);
	glVertex3f (x+d, y+d, z-d);

	glVertex3f (x-d, y+d, z+d);
	glVertex3f (x-d, y-d, z+d);
	glVertex3f (x-d, y-d, z-d);
	glVertex3f (x-d, y+d, z-d);

	glEnd ();
}


/* this function will be called to set up the GL system once, before
   anything can be drawn */

void surf3d_initgl ()
{

	dp (3, "        Setting gl defaults for 3D surface plot\n");

	/* enable depth testing (z-buffering): this means that if we try and draw 
       something behind something already drawn then only the 
       visible parts get drawn */
	glEnable (GL_DEPTH_TEST); 

	glShadeModel(GL_SMOOTH);				/* Gouraud shading in polygons */

	glPolygonMode (GL_FRONT, GL_FILL);		/* Fill forward-facing polygons */
	glPolygonMode (GL_BACK, GL_POINTS);		/* Don't Fill backward-facing polygons */
	
	glEnable(GL_BLEND);							/* Turn on blending (for transparency) */
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	glEnable (GL_NORMALIZE);		/* allow internal normalization of normal vectors */
	
}



/* This function sets up the transformation matrices */

void surf3d_view_init ()
{
float eye_x, eye_y, eye_z, eye_r;
float azimuth, elevation;

	dp (3, "        Initializing gl observer\n");
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity ();		/* We MUST start the computation chain with the identity matrix */

	gluPerspective(surf3d_perspective, /* viewing angle */
		   1.0, /* aspect ratio */
		   1.0, /* min depth (front clipping plane) */
		   12.0); /* max depth (rear clipping plane) */

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity ();

	azimuth = PI*surf3d_azimuth_adj->value / 180.0;
	elevation = PI*surf3d_elev_adj->value / 180.0;
	eye_r = surf3d_eyedist;		/* Distance from center of object */
	eye_z = eye_r*sin(elevation);
	eye_x = eye_r*cos(elevation)*cos(azimuth);
	eye_y = eye_r*cos(elevation)*sin(azimuth);

	gluLookAt(eye_x, eye_y, eye_z,   /* 'eye' location */
	      0.0, 0.0, 0.0,   /* Where we are looking - always the origin */
	      0.0, 0.0, 1.0);  /* up is in positive z direction */

}

/* If lighting is requested, this function sets up the lights */

void surf3d_setup_lights ()
{
GLfloat light_amb[4] = {0.3, 0.3, 0.7, 1.0};
GLfloat light_dif[4] = {1.0, 1.0, 1.0, 1.0};
GLfloat lightv[4];
GLfloat zerov[4] = { 0.0, 0.0, -1.0, 0.0 };

	/* Light0 is both ambient and point source. The position is computed
	   from elevation and azimuth starting at the current (translated)
	   center position by usong the vector (0,0,2) and rotating it
	   around Z for azimuth and arounf X for elevation.
	   A non-lighted red cube is drawn at the light position */

	dp (3, "        Initializing gl lighting\n");
	glPushMatrix ();	
	glRotatef (surf3d_lv_azim, 0.0, 0.0, 1.0);
	glRotatef (90-surf3d_lv_elev, 1.0, 0.0, 0.0);
	lightv[0]=0.0; lightv[1]=0.0; lightv[2]=surf3d_lvdist; 
	lightv[3]=surf3d_spotlight;	/* Zero for directional, one for positional */

	if (surf3d_showlight)		/* Draw a small box at light source position */
	{
		glDisable (GL_LIGHTING);
		cube_at (lightv[0], lightv[1], lightv[2]);
		/*	glBegin (GL_LINES);
			glVertex3fv (lightv);
			glVertex3fv (zerov);
			glEnd ();				*/
	}
	glEnable (GL_LIGHTING);
	glLightfv (GL_LIGHT0, GL_POSITION, lightv);
	if (surf3d_spotlight) 
		glLightf  (GL_LIGHT0, GL_SPOT_CUTOFF, 90.0);
	else
		glLightf  (GL_LIGHT0, GL_SPOT_CUTOFF, 180.0);	/* Convert back into directional */
	if (surf3d_spotlight) glLightfv (GL_LIGHT0, GL_SPOT_DIRECTION, zerov);
	
	glPopMatrix ();
	

	/* Set up the lamp to produce white light for diffuse and specular
	   reflection and a bluish tint for ambient light */

	glLightfv (GL_LIGHT0, GL_AMBIENT, light_amb);
	glLightfv (GL_LIGHT0, GL_DIFFUSE, light_dif);
	glLightfv (GL_LIGHT0, GL_SPECULAR, light_dif);
	glEnable (GL_LIGHT0);
	
	/* Shininess is used globally on all surfaces.
	   This is the cos**n factor. We also use white specular
	   reflection all over the place. */
	   
	glMaterialf (GL_FRONT_AND_BACK, GL_SHININESS, 10.0);
	glMaterialfv (GL_FRONT_AND_BACK, GL_SPECULAR, light_dif);
	
	/* Finally, select local viewer light model */
	
	glLightModeli (GL_LIGHT_MODEL_LOCAL_VIEWER, 1);

}

/* Draw a phantom for testing */

void surf3d_draw_phantom ()
{
GLfloat matv[4];							/* Universal material or light vector */
GLfloat s3 = 1.0 / sqrt(3.0);

	matv[0]=0.2; matv[1]=0.2; matv[2]=0.2; matv[3]=1.0; 

	glBegin(GL_QUADS);

	glColor3f (0.2,0.2,0.2);	/* A dark gray rectangle in the X-Z plane */
	glNormal3f (0.0, 1.0, 0.0);	/* Has a surface normal parallel to y axis */
	if (light_en) glMaterialfv (GL_FRONT, GL_DIFFUSE, matv);
  	glVertex3f ( 2.0, -2.0, 0.0);		/* Must be drawn with vertices CW - why?? */
  	glVertex3f (-2.0, -2.0, 0.0);
  	glVertex3f (-2.0,  2.0, 0.0);
  	glVertex3f ( 2.0,  2.0, 0.0);

	glEnd ();
	
	glBegin (GL_LINES);		/* a coordinate system with 3 axes: */
	
	glColor3f (1,0,0);		/* X-axis from origin to x=3 (red) */
	glVertex3d (3,0,0);
	glVertex3d (0,0,0);
	
	glColor3f (0,1,0);		/* Y-axis from origin to y=3 (green) */
	glVertex3d (0,3,0);
	glVertex3d (0,0,0);
	
	glColor3f (0,0,1);		/* Z-axis from origin to z=3 (blue) */
	glVertex3d (0,0,3);
	glVertex3d (0,0,0);
	
	glEnd ();

	/* Now draw the triangles - draw all vertices CCW */
	glBegin(GL_TRIANGLES);

	matv[0]=1.0; matv[1]=1.0; matv[2]=0.0; matv[3]=1.0; 

	glColor3f (1,1,0);		/* Triangle between x and y */
	glNormal3f (0.0, 0.0, -1.0);	
	if (light_en) glMaterialfv (GL_FRONT, GL_DIFFUSE, matv);
  	glVertex3d (1,0,0);
  	glVertex3d (0,0,0);
  	glVertex3d (0,1,0);

	matv[0]=1.0; matv[1]=0.0; matv[2]=1.0; matv[3]=1.0; 

	glColor3f (1,0,1);		/* Triangle between x and z */
	glNormal3f (0.0, -1.0, 0.0);	
	if (light_en) glMaterialfv (GL_FRONT, GL_DIFFUSE, matv);
  	glVertex3d (0,0,0);
  	glVertex3d (1,0,0);
  	glVertex3d (0,0,1);

	matv[0]=0.0; matv[1]=1.0; matv[2]=1.0; matv[3]=1.0; 

	glColor3f (0,1,1);		/* Triangle between y and z */
	glNormal3f (-1.0, 0.0, 0.0);	
	if (light_en) glMaterialfv (GL_FRONT, GL_DIFFUSE, matv);
  	glVertex3d (0,0,0);
  	glVertex3d (0,0,1);
  	glVertex3d (0,1,0);

	matv[0]=1.0; matv[1]=1.0; matv[2]=1.0; matv[3]=1.0; 

	glColor3f (1,1,1);
	glNormal3f (s3,s3,s3);	
	if (light_en) glMaterialfv (GL_FRONT, GL_DIFFUSE, matv);
  	glVertex3d (1,0,0);
  	glVertex3d (0,1,0);
  	glVertex3d (0,0,1);

	glEnd ();
	
	glFlush ();
}


/* Prepare the vertices from the image data. This function gets called
   for any new image, and then only if subsampling or z-scale has changed */


void surf3d_prep ()
{
float zfac, zscale, min,max,xf;
float v1[3], v2[3], v0[3];						/* the vertices as arrays */
float n1[3], n2[3];								/* The normal vectors */
int x,y,intvl, vxmax, vymax;
int x0,x1,xn1,y0,y1,yn1;
float z;
rgbftriplet rgb;
int idx;

	imgminmax (&surf3d_img, &min, &max);
	zfac = 1.0/(max-min);	/* Max Z scale factor for maximum height == 1 */
	zscale = surf3d_zscale_adj->value/100.0;	/* Reduce z scale with slider value */
	intvl = 1<<((int)(surf3d_subsamp_adj->value)-1);	/* Subsampling interval */
	xf = 0.5*MAX(surf3d_img.xmax, surf3d_img.ymax);	/* x/y scale factor to keep in unit cube */
	vxmax = surf3d_img.xmax / intvl;
	vymax = surf3d_img.ymax / intvl;
	surf3d_vxmax=vxmax; surf3d_vymax=vymax;
	
	dp (3, "        Surface mesh recomputation. Image size (%d by %d)\n",surf3d_img.xmax, surf3d_img.ymax);
	dp (3, "            Downsampled by factor %d to new size (%d by %d)\n",intvl,vxmax,vymax);

	/* Allocate memory if needed */

	if (surf3d_vert) free (surf3d_vert);
	if (surf3d_color) free (surf3d_color);
	if (surf3d_norm) free (surf3d_norm);

	surf3d_vert = calloc (3*vxmax*vymax, sizeof(GLfloat));
	surf3d_color = calloc (4*vxmax*vymax, sizeof(GLfloat));
	surf3d_norm = calloc (3*vxmax*vymax, sizeof(GLfloat));

	/* First, create a list of vertices (including their color and
	   normal vectors). The normal vector is the average of the
	   normals of 6 adjoining triangles. This type of computation is 
	   quite ineffective, but easy to implement. We accept this,
	   because the azimuth/elevation operations and scene render
	   operations do not require recomputation of this section */

	for (y=0; y<vymax; y++)
		for (x=0; x<vxmax; x++)
		{
			idx = x + vxmax*y;
			x0 = x*intvl; y0 = y*intvl;
			x1 = x0+intvl; xn1 = x0-intvl;	/* The neighbor points */
			y1 = y0+intvl; yn1 = y0-intvl;

			z = (readbuf_flt (&surf3d_img,x0,y0,cur_slice)-min)*zfac;
			
			/* The vertex and its color */
			
			v0[0]=surf3d_vert[3*idx  ]=x0/xf-1.0;
			v0[1]=surf3d_vert[3*idx+1]=y0/xf-1.0;
			v0[2]=surf3d_vert[3*idx+2]=z*zscale;
			rgb=surf3d_clut (z);
			surf3d_color[4*idx  ]=rgb.red;
			surf3d_color[4*idx+1]=rgb.green;
			surf3d_color[4*idx+2]=rgb.blue;
			surf3d_color[4*idx+3]=1.0;

			/* The normal vector, computed by averaging the normals
			   of all six adjoining triangles. v0 is the central vertex.
			   Note that we compute v0 - v1 - v2 to be oriented CCW.
			   Note also that we *require* GL to normalize the normal vector,
			   but we normalize the individual components to have equal weights */

			n1[0]=n1[1]=n1[2]=0.0;	/* Reset accumulating vector */

			/* Forward triangle */

			if ((x1 < surf3d_img.xmax) && (y1 < surf3d_img.ymax))
			{
				v1[0] = x1/xf-1.0;
				v1[1] = v0[1];
				v1[2] = (readbuf_flt (&surf3d_img,x1,y0,cur_slice)-min)*zfac*zscale;
				v2[0] = v0[0];
				v2[1] = y1/xf-1.0;
				v2[2] = (readbuf_flt (&surf3d_img,x0,y1,cur_slice)-min)*zfac*zscale;
				crossprodf2 (v0,v1,v2,n2);
				n1[0]+=n2[0]; n1[1]+=n2[1]; n1[2]+=n2[2];
			}

			/* Move to the next triangles clockwise (actually, a quad) */

			if ((x1 < surf3d_img.xmax) && (yn1 <= 0))
			{
				v1[0] = x1/xf-1.0;
				v1[1] = yn1/xf-1.0;
				v1[2] = (readbuf_flt (&surf3d_img,x1,yn1,cur_slice)-min)*zfac*zscale;
				v2[0] = v1[0];
				v2[1] = v0[1];
				v2[2] = (readbuf_flt (&surf3d_img,x1,y0,cur_slice)-min)*zfac*zscale;
				crossprodf2 (v0,v1,v2,n2);
				n1[0]+=n2[0]; n1[1]+=n2[1]; n1[2]+=n2[2];

				v1[0] = v0[0];
				v1[1] = yn1/xf-1.0;
				v1[2] = (readbuf_flt (&surf3d_img,x0,yn1,cur_slice)-min)*zfac*zscale;
				v2[0] = x1/xf-1.0;
				v2[1] = v1[1];
				v2[2] = (readbuf_flt (&surf3d_img,x1,yn1,cur_slice)-min)*zfac*zscale;
				crossprodf2 (v0,v1,v2,n2);
				n1[0]+=n2[0]; n1[1]+=n2[1]; n1[2]+=n2[2];
			}

			/* Backward triangle */

			if ((xn1 >= 0) && (yn1 >= 0))
			{
				v1[0] = xn1/xf-1.0;
				v1[1] = v0[1];
				v1[2] = (readbuf_flt (&surf3d_img,xn1,y0,cur_slice)-min)*zfac*zscale;
				v2[0] = v0[0];
				v2[1] = yn1/xf-1.0;
				v2[2] = (readbuf_flt (&surf3d_img,x0,yn1,cur_slice)-min)*zfac*zscale;
				crossprodf2 (v0,v1,v2,n2);
				n1[0]+=n2[0]; n1[1]+=n2[1]; n1[2]+=n2[2];
			}

			/* and the final quad (two triangles) */

			if ((xn1 >=0) && (y1 < surf3d_img.ymax))
			{
				v1[0] = xn1/xf-1.0;
				v1[1] = y1/xf-1.0;
				v1[2] = (readbuf_flt (&surf3d_img,xn1,y1,cur_slice)-min)*zfac*zscale;
				v2[0] = v1[0];
				v2[1] = v0[1];
				v2[2] = (readbuf_flt (&surf3d_img,xn1,y0,cur_slice)-min)*zfac*zscale;
				crossprodf2 (v0,v1,v2,n2);
				n1[0]+=n2[0]; n1[1]+=n2[1]; n1[2]+=n2[2];

				v1[0] = v0[0];
				v1[1] = y1/xf-1.0;
				v1[2] = (readbuf_flt (&surf3d_img,x0,y1,cur_slice)-min)*zfac*zscale;
				v2[0] = xn1/xf-1.0;
				v2[1] = v1[1];
				v2[2] = (readbuf_flt (&surf3d_img,xn1,y1,cur_slice)-min)*zfac*zscale;
				crossprodf2 (v0,v1,v2,n2);
				n1[0]+=n2[0]; n1[1]+=n2[1]; n1[2]+=n2[2];
			}

			surf3d_norm[3*idx  ]=n1[0];
			surf3d_norm[3*idx+1]=n1[1];
			surf3d_norm[3*idx+2]=n1[2];

		}
	
	/* Done. For each subsampled pixel, we now have one vertex and one normal. */
	surf3d_prepared=1;
	
}

			



/* Draw the triangles from the list */

void surf3d_redraw_trig ()
{
GLfloat matv[4];							/* Universal material or light vector */
GLfloat mat_spec[4];
int x,y,idx;
float shine;

	if (!surf3d_prepared) surf3d_prep ();

	dp (3, "        Redrawing entire surface (core function)\n");

	/* The shadow of the base area. A unit-size rectangle in the x-y plane
	   at z=0. Dark gray color and totally diffuse, i.e. specular=0 */

	matv[0]=matv[1]=matv[2]=0.2; matv[3]=1.0;
	mat_spec[0] = mat_spec[1] = mat_spec[2] = 0.0; mat_spec[3]=1.0;

	glBegin(GL_QUADS);

	glColor4fv (matv);	/* A dark gray rectangle in the X-Y plane */
	glNormal3f (0.0, 0.0, 1.0);	/* Has a surface normal parallel to z axis */
	if (light_en) glMaterialfv (GL_FRONT, GL_AMBIENT_AND_DIFFUSE, matv);
	if (light_en) glMaterialfv (GL_FRONT, GL_SPECULAR, mat_spec);
  	glVertex3f ( 1.0, -1.0, 0.0);		/* Must be drawn with vertices CW - why?? */
  	glVertex3f (-1.0, -1.0, 0.0);
  	glVertex3f (-1.0,  1.0, 0.0);
  	glVertex3f ( 1.0,  1.0, 0.0);

	glEnd ();

	/* Compute the (white!) specular reflection intensity */
	shine = surf3d_lv_spec / 100.0;
	mat_spec[0] = mat_spec[1] = mat_spec[2] = shine; mat_spec[3]=1.0;
	glMaterialfv (GL_FRONT_AND_BACK, GL_SPECULAR, mat_spec);

	/* Enable array processing for vertices, color and normals,
	   then hand over the pointers */
	
	glEnableClientState (GL_VERTEX_ARRAY);
	glEnableClientState (GL_COLOR_ARRAY);
	glEnableClientState (GL_NORMAL_ARRAY);
	glVertexPointer (3, GL_FLOAT, 0, surf3d_vert);
	glColorPointer (4, GL_FLOAT, 0, surf3d_color);
	glNormalPointer (GL_FLOAT, 0, surf3d_norm);

	/* Transmit all vertices. We need to use glArrayElement rather than
	   glDrawElements because we need to specify material color for
	   each vertex anyway.
	   By using glColorMaterial we copy the color buffer into the 
	   material property buffer */

	glEnable (GL_COLOR_MATERIAL);
	glColorMaterial (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

	glBegin (GL_TRIANGLES);
	
	for (y=0; y<surf3d_vymax-1; y++)
		for (x=0; x<surf3d_vxmax-1; x++)
		{
			idx = x + surf3d_vxmax*y;

			glArrayElement (idx);
			glArrayElement (idx+1);
			glArrayElement (idx+surf3d_vxmax);

			glArrayElement (idx+1);
			glArrayElement (idx+1+surf3d_vxmax);
			glArrayElement (idx+surf3d_vxmax);
			
		}
	
	glEnd ();
	
	glDisable (GL_COLOR_MATERIAL);

	glDisableClientState (GL_VERTEX_ARRAY);
	glDisableClientState (GL_COLOR_ARRAY);
	glDisableClientState (GL_NORMAL_ARRAY);

	glFlush ();

}


/****************************** !!! ***********************************

	Below this point, we encounter mixed gdl/gl commands.

*/






/* Redraw it all, including new view point */

void surf3d_redraw ()
{
	dp (3, "        GL surface plot redraw\n");

	/* gtk_widget_grab_focus(GTK_WIDGET(glsurfarea)); */

	/* 0.9.0 changes to gl-ext, and gtk_gl_area_make_current is now deprecated, too */
	//if (gtk_gl_area_make_current(GTK_GL_AREA(glsurfarea))) 


	glViewport(0,0, glsurfarea->allocation.width, glsurfarea->allocation.height);
	if (light_en) 
		glEnable (GL_LIGHTING);
	else
		glDisable (GL_LIGHTING);

		/* Prepare the clear command by specifying how to clear */

	switch (surf3d_bgcolor)		/* Set clear value for color buffer = bkgnd color */
	{
		case 0:	glClearColor(0.0,0.0,0.0,0.0);
				break;
		case 1:	glClearColor(0.3,0.3,0.3,0.0);
				break;
		case 2:	glClearColor(0.6,0.6,0.6,0.0);
				break;
		case 3:	glClearColor(1.0,1.0,1.0,0.0);
				break;
		default:
				glClearColor(0.0,0.0,0.0,0.0);
	}		
	glClearDepth (1.0);

	glDrawBuffer(GL_FRONT_AND_BACK);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	surf3d_view_init ();		/* Viewport, observer's position etc */
	if (light_en) surf3d_setup_lights();
	
	/*surf3d_draw_phantom (); */
	surf3d_redraw_trig ();			/* Triangles. Lots of them. */

}



gint glsurfarea_configure (GtkWidget *widget, GdkEventConfigure *event)
{
  
	dp (2, "GL surface plot configure event\n");
	
	gldrawable = gtk_widget_get_gl_drawable (glsurfarea);
	glcontext = gtk_widget_get_gl_context (glsurfarea);

	if (gdk_gl_drawable_gl_begin (gldrawable, glcontext))			/* For gl-ext in 0.09. */
	{
		dp (3, "     ----------> gdk_gl_drawable_gl_begin\n");
    	glViewport(0,0, widget->allocation.width, widget->allocation.height);
		surf3d_initgl ();
		dp (3, "     ----------> gdk_gl_drawable_gl_end\n");
		gdk_gl_drawable_gl_end (gldrawable);
	}
	
	return TRUE;
}


gint glsurfarea_expose (GtkWidget *widget, GdkEventExpose *event)
{
  
	if(event->count > 0) /* Draw only last expose */
	{
		return TRUE;
	}

	dp (2, "GL surface plot expose event (filtered)\n");

	gldrawable = gtk_widget_get_gl_drawable (glsurfarea);
	glcontext = gtk_widget_get_gl_context (glsurfarea);

	if (gdk_gl_drawable_gl_begin (gldrawable, glcontext))			/* For gl-ext in 0.09. */
	{
		dp (3, "     ----------> gdk_gl_drawable_gl_begin\n");
		surf3d_redraw ();
		if (gdk_gl_drawable_is_double_buffered (gldrawable))
			gdk_gl_drawable_swap_buffers (gldrawable);
		else
			glFlush ();
		dp (3, "     ----------> gdk_gl_drawable_gl_end\n");
		gdk_gl_drawable_gl_end (gldrawable);
	}

	return TRUE;
}


void gl_area_force_redraw ()
{
int w,h;

	w = glsurfarea->allocation.width;
	h = glsurfarea->allocation.height;

	gtk_widget_queue_draw_area (glsurfarea, 0, 0, w, h);
	
}

/****************************************************+

	GL Surface plot GUI and gtk event handlers
	
*****************************************************/

GtkAdjustment* surf3d_lv_elev_adj;
GtkAdjustment* surf3d_lv_azim_adj;
GtkAdjustment* surf3d_lv_spec_adj;
GtkAdjustment* surf3d_eyedist_adj;
GtkAdjustment* surf3d_lvdist_adj;
GtkWidget* surf3d_advcheck;


void surf3d_go ()
{

	surf3d_lv_elev = surf3d_lv_elev_adj->value;
	surf3d_lv_azim = surf3d_lv_azim_adj->value;
	surf3d_lv_spec = surf3d_lv_spec_adj->value;
	surf3d_eyedist = surf3d_eyedist_adj->value;
	surf3d_lvdist = surf3d_lvdist_adj->value;
	surf3d_perspective = surf3d_perspective_adj->value;
	gl_area_force_redraw ();

}


void on_surf3d_slider_changed (GtkWidget* widget, gpointer user_data)
{
int slidernum;

	slidernum= *(int*)user_data;
	dp (3, "3D surface plot: Slider %d changed\n",slidernum);
	if ((slidernum==menuopt3) || (slidernum==menuopt4))
	{
		surf3d_prepared=0;	/* Force recomputation of vertex field */
	}
	if (surf3d_awnd_exists) 
		surf3d_go ();
	else
		gl_area_force_redraw ();
}


void on_surf3d_advslider_changed (GtkWidget* widget, gpointer user_data)
{
int slidernum;

	slidernum= *(int*)user_data;
	dp (3, "3D surface plot: Advanced slider %d changed\n",slidernum);
	surf3d_go ();
}

void destroy_surf3d_awnd (GtkWidget *widget, gpointer user_data)
{
	surf3d_awnd_exists = -1;
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (surf3d_advcheck), 0);
	surf3d_awnd_exists = 0;
	gtk_widget_destroy ( GTK_WIDGET ( surf3d_awnd ) );
}


void destroy_surf3dwnd (GtkWidget *widget, gpointer user_data)
{

	if (surf3d_vert) free (surf3d_vert);
	if (surf3d_color) free (surf3d_color);
	if (surf3d_norm) free (surf3d_norm);
	surf3d_vert=NULL; surf3d_color=NULL; surf3d_norm=NULL;
	
	if (surf3d_awnd_exists) destroy_surf3d_awnd (NULL, NULL);
	surf3d_wnd_exists = 0;
	
	/* gtk_widget_destroy ( GTK_WIDGET ( surf3dwnd ) ); */
	gtk_exit (0);
}



void on_surf3d_set_lightmenu (GtkWidget *widget, gpointer user_data)
{
int function;

	function = *(int*) user_data;
	
	if (function==0)
		light_en=0;
	else
	{
		light_en=1;
		if (function==1) surf3d_spotlight=0;
		if (function==2) surf3d_spotlight=1;
	}
	surf3d_go ();

}


void on_surf3d_set_bgmenu (GtkWidget *widget, gpointer user_data)
{

	surf3d_bgcolor = *(int*) user_data;
	surf3d_go ();

}

void on_surf3d_view_lsource (GtkToggleButton *togglebutton, gpointer user_data)
{
	surf3d_showlight = !surf3d_showlight;
	surf3d_go ();
}


void lut_optionmenu_handler (GtkWidget *widget, gpointer user_data)
{
int i;

	i = *(int*) user_data;
	dp (2, "Switched to LUT %d\n", i);
	current_lut = i;
	surf3d_prepared=0;	/* Force recomputation of vertex field for recoloring */
	surf3d_go ();

}



void on_surf3d_save_clicked (GtkWidget *widget, gpointer user_data)
{
int w,h, success;
GdkPixbuf *pixbuf;
rgbtriplet* tempimg;
char msg[1024];
gchar* selectedfile;
gchar *fmt;
GError *error = NULL;

	/* Get the necessary memory for temporary frame storage.
	   The deviation over tempimg is necessary because gdk_imlib
	   does not have a blank image allocation function, so we need
	   to copy the data twice */

	w = glsurfarea->allocation.width; h = glsurfarea->allocation.height;
	tempimg = (rgbtriplet*) calloc (w*h, sizeof (rgbtriplet));
	if (!tempimg) fatal_memerr ();

	gldrawable = gtk_widget_get_gl_drawable (glsurfarea);
	glcontext = gtk_widget_get_gl_context (glsurfarea);
	if (gdk_gl_drawable_gl_begin (gldrawable, glcontext))			/* For gl-ext in 0.09. */
	{
 		glReadPixels (0,0,w,h,GL_RGB,GL_UNSIGNED_BYTE, tempimg);
		gdk_gl_drawable_gl_end (gldrawable);
	}

	/* Copy the data into the pixbuf */

	pixbuf = gdk_pixbuf_new_from_data ((unsigned char*)tempimg, GDK_COLORSPACE_RGB, FALSE, 8, 
														w, h, 3*w /* rowstride */,
														pixbuf_destroy_handler /* destroy callback */,
														&pixbuf /* destroy callback user_data */);


	/* Let the user choose a file name and save the file using pixbuf */

	ChooseFile ("Save scene as (suffix determines type):", widget, 1, &selectedfile);
	if (selectedfile)
	{
		fmt = strrchr(selectedfile, '.');
		if (fmt)
			fmt++;
		else
			strcpy (fmt, "png");			/* Defaults to PNG */

		success = gdk_pixbuf_save (pixbuf, selectedfile, fmt, &error, NULL);
		if (!success)
		{
			sprintf (msg, "Save operation failed.\n%s", error->message);
			ShowMessage (msg);
		}
		else
			statusbar_display ("Scene image saved.");
	}
	g_object_unref (G_OBJECT(pixbuf));
}


/* Advanced controls */

void surf3d_advanced ()
{
GtkWidget *vbox, *table, *hs1;
GtkWidget *label0, *label1, *label2, *label3, *label4, *label5, *label6, *label7, *label8, *label9;
GtkWidget *Hsurf3d_lv_elev, *Hsurf3d_lv_azim, *Hsurf3d_lv_spec, *Hsurf3d_lvdist, *Hsurf3d_eyedist, *Hsurf3d_perspective;
GtkWidget *surf3d_lightmenu;
GtkWidget *surf3d_lightmenu_menu;
GtkWidget *surf3d_bgmenu;
GtkWidget *surf3d_bgmenu_menu;
GtkWidget* menuitem;
GtkWidget* menuitem1;
GtkWidget *predef_lut;
GtkWidget *predef_lut_menu;
GtkWidget *showlight_toggle, *swbutton3;
int y;


	surf3d_awnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (surf3d_awnd), "surf3d_awnd", surf3d_awnd);
	gtk_window_set_title (GTK_WINDOW (surf3d_awnd), "GL 3D Advanced Controls");

	vbox = vbox_in_window (surf3d_awnd, "s3dvbox2");
	gtk_container_border_width(GTK_CONTAINER(vbox), 5);
	table = MakeTable (surf3d_awnd,vbox,"stab2",2,10);
	y=0;

	label0 = MakeLabel (surf3d_awnd,table,"Observer distance","s3d_obsd",0,y);
	surf3d_eyedist_adj = GTK_ADJUSTMENT(gtk_adjustment_new (surf3d_eyedist, 1, 10, 0.1, 1, 0));
	Hsurf3d_eyedist = MakeHScale (surf3d_eyedist_adj, surf3d_awnd, table, "s3d_hscale9", 1,y);
	y++;

	label9 = MakeLabel (surf3d_awnd,table,"Perspective","s3d_pspct",0,y);
	surf3d_perspective_adj = GTK_ADJUSTMENT(gtk_adjustment_new (surf3d_perspective, 10, 60, 1, 10, 0));
	Hsurf3d_perspective = MakeHScale (surf3d_perspective_adj, surf3d_awnd, table, "s3d_hscale10", 1,y);
	y++;

	label7 = MakeLabel (surf3d_awnd,table,"Background color","s3d_lbg",0,y);
	surf3d_bgmenu = gtk_option_menu_new ();
	gtk_widget_ref (surf3d_bgmenu);
	gtk_object_set_data_full (GTK_OBJECT (surf3d_awnd), "surf3d_menu2", surf3d_bgmenu,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (surf3d_bgmenu);
	gtk_table_attach (GTK_TABLE (table), surf3d_bgmenu, 1, 2, y, y+1,
                      (GtkAttachOptions) (0),
                      (GtkAttachOptions) (0), 0, 0);

	surf3d_bgmenu_menu = gtk_menu_new ();
	menuitem = Option_menuitem (surf3d_bgmenu_menu, GTK_SIGNAL_FUNC(on_surf3d_set_bgmenu), "Black", (gpointer) &menuopt0);
	menuitem = Option_menuitem (surf3d_bgmenu_menu, GTK_SIGNAL_FUNC(on_surf3d_set_bgmenu), "Dark grey", (gpointer) &menuopt1);
	menuitem = Option_menuitem (surf3d_bgmenu_menu, GTK_SIGNAL_FUNC(on_surf3d_set_bgmenu), "Light grey", (gpointer) &menuopt2);
	menuitem = Option_menuitem (surf3d_bgmenu_menu, GTK_SIGNAL_FUNC(on_surf3d_set_bgmenu), "White", (gpointer) &menuopt3);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (surf3d_bgmenu), surf3d_bgmenu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (surf3d_bgmenu), surf3d_bgcolor);
	gtk_container_set_border_width (GTK_CONTAINER (surf3d_bgmenu), 5);
	y++;

	surf3d_lightmenu = gtk_option_menu_new ();
	gtk_widget_ref (surf3d_lightmenu);
	gtk_object_set_data_full (GTK_OBJECT (surf3d_awnd), "surf3d_menu1", surf3d_lightmenu,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (surf3d_lightmenu);
	gtk_table_attach (GTK_TABLE (table), surf3d_lightmenu, 1, 2, y, y+1,
                      (GtkAttachOptions) (0),
                      (GtkAttachOptions) (0), 0, 0);

	label1 = MakeLabel (surf3d_awnd,table,"Enable Lighting","s3d_lgt",0,y);
	surf3d_lightmenu_menu = gtk_menu_new ();
	menuitem = Option_menuitem (surf3d_lightmenu_menu, GTK_SIGNAL_FUNC(on_surf3d_set_lightmenu), "Unlit scene", (gpointer) &menuopt0);
	menuitem = Option_menuitem (surf3d_lightmenu_menu, GTK_SIGNAL_FUNC(on_surf3d_set_lightmenu), "Directional light", (gpointer) &menuopt1);
	menuitem = Option_menuitem (surf3d_lightmenu_menu, GTK_SIGNAL_FUNC(on_surf3d_set_lightmenu), "Spotlight", (gpointer) &menuopt2);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (surf3d_lightmenu), surf3d_lightmenu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (surf3d_lightmenu), 0);
	gtk_container_set_border_width (GTK_CONTAINER (surf3d_lightmenu), 5);
	y++;
	
	label2 = MakeLabel (surf3d_awnd,table,"Light vector azimuth","s3d_lva",0,y);
	surf3d_lv_azim_adj = GTK_ADJUSTMENT(gtk_adjustment_new (surf3d_lv_azim, -180, 180, 1, 10, 1));
	Hsurf3d_lv_azim = MakeHScale (surf3d_lv_azim_adj, surf3d_awnd, table, "s3d_hscale6", 1,y);
	y++;

	label3 = MakeLabel (surf3d_awnd,table,"Light vector elevation","s3d_lve",0,y);
	surf3d_lv_elev_adj = GTK_ADJUSTMENT(gtk_adjustment_new (surf3d_lv_elev, 0, 90, 1, 10, 1));
	Hsurf3d_lv_elev = MakeHScale (surf3d_lv_elev_adj, surf3d_awnd, table, "s3d_hscale5", 1,y);
	y++;

	label4 = MakeLabel (surf3d_awnd,table,"Specular intensity","s3d_lvs",0,y);
	surf3d_lv_spec_adj = GTK_ADJUSTMENT(gtk_adjustment_new (surf3d_lv_spec, 0, 100, 1, 10, 1));
	Hsurf3d_lv_spec = MakeHScale (surf3d_lv_spec_adj, surf3d_awnd, table, "s3d_hscale7", 1,y);
	y++;

	label5 = MakeLabel (surf3d_awnd,table,"Light source distance","s3d_lvd",0,y);
	surf3d_lvdist_adj = GTK_ADJUSTMENT(gtk_adjustment_new (surf3d_lvdist, 1, 20, 0.1, 1, 0));
	Hsurf3d_lvdist = MakeHScale (surf3d_lvdist_adj, surf3d_awnd, table, "s3d_hscale8", 1,y);
	y++;

	label6 = MakeLabel (surf3d_awnd,table,"Show light source","s3d_lls",0,y);
	showlight_toggle = Checkbutton_in_table (surf3d_awnd, table, " ", "showlight_tog", 1,y, on_surf3d_view_lsource);
	y++;

	label8 = MakeLabel (surf3d_awnd,table,"Color LUT","s3d_lut",0,y);
	predef_lut = Make_Optionmenu (surf3d_awnd,table,"predef_lut",1,y);
	predef_lut_menu = gtk_menu_new ();
	menuitem1 = Option_menuitem (predef_lut_menu, GTK_SIGNAL_FUNC(lut_optionmenu_handler), "Default", (gpointer) &menuopt0);
	menuitem1 = Option_menuitem (predef_lut_menu, GTK_SIGNAL_FUNC(lut_optionmenu_handler), "Grayscale", (gpointer) &menuopt1);
	menuitem1 = Option_menuitem (predef_lut_menu, GTK_SIGNAL_FUNC(lut_optionmenu_handler), "Negative", (gpointer) &menuopt2);
	menuitem1 = Option_menuitem (predef_lut_menu, GTK_SIGNAL_FUNC(lut_optionmenu_handler), "RGB simple", (gpointer) &menuopt3);
	menuitem1 = Option_menuitem (predef_lut_menu, GTK_SIGNAL_FUNC(lut_optionmenu_handler), "RGB w/black", (gpointer) &menuopt4);
	menuitem1 = Option_menuitem (predef_lut_menu, GTK_SIGNAL_FUNC(lut_optionmenu_handler), "NIH Fire", (gpointer) &menuopt5);
	menuitem1 = Option_menuitem (predef_lut_menu, GTK_SIGNAL_FUNC(lut_optionmenu_handler), "Rainbow", (gpointer) &menuopt6);
	menuitem1 = Option_menuitem (predef_lut_menu, GTK_SIGNAL_FUNC(lut_optionmenu_handler), "Gamma 2.0", (gpointer) &menuopt10);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (predef_lut), predef_lut_menu);
	y++;


	hs1 = MakeHsep (surf3d_awnd,vbox, "shsep3",-1,-1);
 	swbutton3 = Button_in_table (surf3d_awnd,vbox,"Dismiss","swbutton3",-1,-1,destroy_surf3d_awnd);


	gtk_signal_connect (GTK_OBJECT (surf3d_eyedist_adj), "value_changed",
                      GTK_SIGNAL_FUNC (on_surf3d_advslider_changed), &menuopt0);
	gtk_signal_connect (GTK_OBJECT (surf3d_lv_elev_adj), "value_changed",
                      GTK_SIGNAL_FUNC (on_surf3d_advslider_changed), &menuopt1);
	gtk_signal_connect (GTK_OBJECT (surf3d_lv_azim_adj), "value_changed",
                      GTK_SIGNAL_FUNC (on_surf3d_advslider_changed), &menuopt2);
 	gtk_signal_connect (GTK_OBJECT (surf3d_lv_spec_adj), "value_changed",
                      GTK_SIGNAL_FUNC (on_surf3d_advslider_changed), &menuopt3);
 	gtk_signal_connect (GTK_OBJECT (surf3d_lvdist_adj), "value_changed",
                      GTK_SIGNAL_FUNC (on_surf3d_advslider_changed), &menuopt4);
  	gtk_signal_connect (GTK_OBJECT (surf3d_perspective_adj), "value_changed",
                      GTK_SIGNAL_FUNC (on_surf3d_advslider_changed), &menuopt5);
 
	gtk_signal_connect (GTK_OBJECT (surf3d_awnd), "destroy",
                    	GTK_SIGNAL_FUNC (destroy_surf3d_awnd),surf3d_awnd);

	surf3d_awnd_exists = 1;
	gtk_widget_show (surf3d_awnd);

}


/* Create the main surface 3D window, including the gtk-gl area */

void on_surf3d_checkbutton1_toggled (GtkToggleButton *togglebutton, gpointer user_data)
{
	if (surf3d_awnd_exists>0)
	{
		surf3d_awnd_exists = 0;
		gtk_widget_destroy ( GTK_WIDGET ( surf3d_awnd ) );
	}
	else
		if (surf3d_awnd_exists==0) surf3d_advanced ();

}


void surf3d_create_window ()
{
GtkWidget *vbox1;
GtkWidget* hbox1;
GtkWidget* table1;
GtkWidget* swbutton1;
GtkWidget* swbutton2;
GtkWidget *hseparator1, *hseparator2;
GtkWidget *surf3d_azimuth, *surf3d_elev, *surf3d_zscale, *surf3d_subsamp;
GtkWidget *label1, *label2, *label3, *label4, *label5;

int submax,subset,subsmax;


	subset=1;
	submax=1; 
	subsmax=1;
	while (submax < MIN(surf3d_img.xmax,surf3d_img.ymax)) 
	{
		submax = submax <<1;
		subsmax++;
	}
	if (submax>8) submax=submax>>3;
	if (MAX(surf3d_img.xmax,surf3d_img.ymax) >=1024)
		subset=4;
	else if (MAX(surf3d_img.xmax,surf3d_img.ymax) >=512)
		subset=3;
	else if (MAX(surf3d_img.xmax,surf3d_img.ymax) >=256)
		subset=2;
		

	surf3dwnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (surf3dwnd), "surf3dwnd", surf3dwnd);
	gtk_window_set_title (GTK_WINDOW (surf3dwnd), "GL 3D Surface Plot");

	/* Items will be placed top to bottom into this vbox as they come */

	vbox1 = vbox_in_window (surf3dwnd, "s3dvbox1");
	gtk_container_border_width(GTK_CONTAINER(vbox1), 5);
	

	/* Start with the GL window: Create new OpenGL widget with GTK's 
	   window mode; analogous to 
	   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH); 
		Changed to gl-ext as of 0.9.0. */
	
	glsurfarea = gtk_drawing_area_new ();
	gtk_widget_set_size_request (glsurfarea, 320, 240); /* minimum size */
	if (gtk_widget_set_gl_capability (glsurfarea, glconfig, NULL, TRUE, GDK_GL_RGBA_TYPE) == 0)
	{
		dp (0, "Error in DSO: Failed creating GL-capable widget\n");
	}

	/* Set up events and signals for OpenGL widget */
	gtk_widget_set_events (GTK_WIDGET(glsurfarea),
                        		GDK_EXPOSURE_MASK|
                        		GDK_BUTTON_PRESS_MASK|
                        		GDK_BUTTON_RELEASE_MASK|
                        		GDK_KEY_PRESS_MASK|
                        		GDK_KEY_RELEASE_MASK|
								GDK_POINTER_MOTION_MASK|
								GDK_POINTER_MOTION_HINT_MASK);
	
	gtk_signal_connect (GTK_OBJECT(glsurfarea), "configure_event", GTK_SIGNAL_FUNC(glsurfarea_configure), NULL);
	gtk_signal_connect (GTK_OBJECT(glsurfarea), "expose_event", GTK_SIGNAL_FUNC(glsurfarea_expose), NULL);

 	gtk_container_add (GTK_CONTAINER(vbox1), glsurfarea);
	gtk_widget_show (glsurfarea);

 
	/* a nice-looking hseparator and a table for the adjustments */

	hseparator1 = MakeHsep (surf3dwnd,vbox1, "shsep1",-1,-1);

	table1 = MakeTable (surf3dwnd,vbox1,"stab1",2,5);
	
	/* The adjustments with labels */
	
	label1 = MakeLabel (surf3dwnd,table1,"Azimuth","s3d_az",0,0);
	label2 = MakeLabel (surf3dwnd,table1,"Elevation","s3d_el",0,1);
	label3 = MakeLabel (surf3dwnd,table1,"Z scale","s3d_zs",0,2);
	label4 = MakeLabel (surf3dwnd,table1,"Subsampling","s3d_ssa",0,3);
	label5 = MakeLabel (surf3dwnd,table1,"Advanced Controls","s3d_adv",0,4);

	surf3d_azimuth_adj = GTK_ADJUSTMENT(gtk_adjustment_new (0, -180, 180, 1, 10, 1));
	surf3d_azimuth = MakeHScale (surf3d_azimuth_adj, surf3dwnd, table1, "s3d_hscale1", 1,0);
	surf3d_elev_adj = GTK_ADJUSTMENT(gtk_adjustment_new (35, 0, 90, 1, 10, 1));
	surf3d_elev = MakeHScale (surf3d_elev_adj, surf3dwnd, table1, "s3d_hscale2", 1,1);
	surf3d_zscale_adj = GTK_ADJUSTMENT(gtk_adjustment_new (20, 1, 100, 1, 10, 1));
	surf3d_zscale = MakeHScale (surf3d_zscale_adj, surf3dwnd, table1, "s3d_hscale3", 1,2);
	surf3d_subsamp_adj = GTK_ADJUSTMENT(gtk_adjustment_new (subset, 1, subsmax, 1, 1, 1));
	surf3d_subsamp = MakeHScale (surf3d_subsamp_adj, surf3dwnd, table1, "s3d_hscale4", 1,3);
	gtk_scale_set_digits (GTK_SCALE (surf3d_subsamp), 0);

	surf3d_advcheck = Checkbutton_in_table (surf3dwnd, table1, " ", "advcheck1", 1,4, on_surf3d_checkbutton1_toggled);
	
	/* a hbox to contain the buttons */

	hseparator2 = MakeHsep (surf3dwnd,vbox1, "shsep2",-1,-1);
	
	hbox1 = gtk_hbox_new (FALSE, 0);
	gtk_widget_show (hbox1);
	/* gtk_container_add (GTK_CONTAINER (vbox1), hbox1); */
	gtk_box_pack_start (GTK_BOX (vbox1), hbox1, FALSE, FALSE, 2);
 
 	swbutton1 = Button_in_table (surf3dwnd,hbox1,"Save scene image","swbutton1",-1,-1,on_surf3d_save_clicked);
 	swbutton2 = Button_in_table (surf3dwnd,hbox1,"Dismiss","swbutton2",-1,-1,destroy_surf3dwnd);

	/* A statusbar (partly for compatibility reasons) */
 
	statusbar1 = gtk_statusbar_new ();
	gtk_widget_ref (statusbar1);
	gtk_object_set_data_full (GTK_OBJECT (surf3dwnd), "statusbar1", statusbar1,
								(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (statusbar1);
	gtk_box_pack_start (GTK_BOX (vbox1), statusbar1, FALSE, FALSE, 0);

	gtk_signal_connect (GTK_OBJECT (surf3d_azimuth_adj), "value_changed",
                      GTK_SIGNAL_FUNC (on_surf3d_slider_changed), &menuopt1);
	gtk_signal_connect (GTK_OBJECT (surf3d_elev_adj), "value_changed",
                      GTK_SIGNAL_FUNC (on_surf3d_slider_changed), &menuopt2);
	gtk_signal_connect (GTK_OBJECT (surf3d_zscale_adj), "value_changed",
                      GTK_SIGNAL_FUNC (on_surf3d_slider_changed), &menuopt3);
	gtk_signal_connect (GTK_OBJECT (surf3d_subsamp_adj), "value_changed",
                      GTK_SIGNAL_FUNC (on_surf3d_slider_changed), &menuopt4);
 
	gtk_signal_connect (GTK_OBJECT (surf3dwnd), "destroy",
                    	GTK_SIGNAL_FUNC (destroy_surf3dwnd),surf3dwnd);


	/* After all is done, grab the focus for glsurfarea and init GL,
	   _then_ show all windows -- doesn`t work properly, so
	   we really need to call initgl in the configure event (!) */

// 	GTK_WIDGET_SET_FLAGS (glsurfarea,GTK_CAN_FOCUS);
// 	gtk_widget_grab_focus (GTK_WIDGET (glsurfarea));
// 	surf3d_initgl();

	surf3d_wnd_exists = 1;
	gtk_widget_show (surf3dwnd);
	gtk_widget_realize (glsurfarea);


}





void surf3d_main ()
{
	
	light_en = 0;
	surf3d_awnd_exists=0;
	surf3d_prepared=0;

	surf3d_perspective=45.0;
	surf3d_lv_elev = 35;
	surf3d_lv_azim = 0;
	surf3d_lv_spec = 60;
	surf3d_eyedist = 4.0;
	surf3d_lvdist = 2.0;
	surf3d_spotlight = 0;
	surf3d_showlight=0;
	surf3d_bgcolor=0;
	
	surf3d_vert=NULL; surf3d_color=NULL; surf3d_norm=NULL; /* Paranoia. They are supposed to be NULL anyway */

	if (glconfig == NULL)
	{
		glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGBA | GDK_GL_MODE_DEPTH | GDK_GL_MODE_DOUBLE);
		if (glconfig == NULL)
		{
		dp (1, "Note: No double-buffered visual available. Trying single-buffered visual\n");
			glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGBA | GDK_GL_MODE_DEPTH);
			if (glconfig == NULL) 
			{
				dp (0,"Error in GL module: Failed creating gl config\n");
				exit ( -1);
			}
		}
	}

	surf3d_create_window ();

}




#define nextargi (--argc,atoi(*++argv))
#define nextargf (--argc,atof(*++argv))
#define nextargs (--argc,*++argv)



int main (int argc, char *argv[])
{
char filename[1024];
char tmpname[1024];
int err, vmaj, vmin;

	for (err=0; err < argc; err++)
		dp (1, "Command-line args: Arg[%d] is %s\n",err,argv[err]);


	dp (1, "Imager 3D Surface Viewer: Initializing GTK\n");
 	gtk_init (&argc, &argv);

	gtk_widget_push_visual   (gdk_rgb_get_visual());
	gtk_widget_push_colormap (gdk_rgb_get_cmap());


	/* Check if OpenGL (GLX extension) is supported. 
	   If not, don't register this module.
		Changed to gl-ext for 0.9.0 */

	dp (0, "Imager 3D Surface Viewer: Checking for pixbuf and GL\n");
	dp (0, "Pixbuf version %s\n", gdk_pixbuf_version);

	gtk_gl_init (&argc, &argv);
	glcontext = NULL;
	gldrawable = NULL;
	if (!gdk_gl_query_extension ())
	{
		dp (0, "gdk_gl_query_extension returned FALSE - no GLX available?\n");
		return -1;
	}
	dp (0, "GLX is available.\n");



	filename[0] = 0;
	tmpname[0] = 0;
	cur_slice=0;
	current_lut = 0;
	gl_widget_ready = 0;

    while (--argc && **++argv=='-') 
	{

		if (!strcmp(argv[0],"-l") || !strcmp(argv[0],"--load"))
			strcpy (filename,nextargs);

		else if (!strcmp(argv[0],"-t") || !strcmp(argv[0],"--tempdir"))
			strcpy (tmpname,nextargs);
		
		else if (!strcmp(argv[0],"-d") || !strcmp(argv[0],"--debuglevel"))
			debuglevel = nextargi;

        else 
		{
        	if (strcmp(argv[0],"--help") && strcmp (argv[0],"-h"))
            	fprintf(stderr,"unknown option '%s'; \n",argv[0]);
			prhelp ();

			exit (1);
		}
	}

	gdk_gl_query_version (&vmaj, &vmin);
	dp (1, "gdk gl-ext version %d.%d is available\n", vmaj, vmin);

	dp (2, "Now attempting to load file:\n");

	if (strlen(filename)<1)
	{
		dp (0, "Error: No file specified. Try -h for help.\n");
		gtk_exit (-1);
	}

	if (strlen(tmpname)>0)
	{
		if (tmpname[strlen(tmpname)-1]!='/')
			strcat (tmpname, "/");
		strncat (tmpname, filename, 1020);
		strcpy (filename, tmpname);
	}
	dp (2, "Complete filename to load: %s\n", filename);

	if ((err=tiffread (&surf3d_img, filename)) < 0)
	{
		dp (0, "Loading file %s failed\n",filename);
		gtk_exit (err);
	}

	mirrory (&surf3d_img);		/* Display is mirrored about Y, so un-mirror it */
	surf3d_main ();



	gtk_main ();
	

	return 0;		/* Dummy exit since we exit through gtk_exit() */
}


/*******************************************************************/




